/* Copyright (c) 2008-09 lvqcl.  All rights reserved.
 *
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 2.1 of the License, or (at
 * your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 */ 

#include "foo_dsp_lpf.h"

const int buf_samples = 4096;

void dsp_lowpass::init()
{
	t_dsp_lpf_params params;
	params.get_lpfconfig(cfg_);

	lpf_state_ = NULL;

	sample_rate_ = 0;
	channel_count_ = 0;
	
	buf_channels_ = 0;
	buffer_ = NULL;

	in_samples_accum_ = out_samples_gen_accum_ = 0;
}

dsp_lowpass::dsp_lowpass()
{
	init();
}

dsp_lowpass::dsp_lowpass(const dsp_preset& p_data)
{
	init();
	t_dsp_lpf_params params(cfg_);
	params.set_data(p_data);
	params.get_lpfconfig(cfg_);
}

/*dsp_lowpass::dsp_lowpass(const dsp_preset& p_data)
{
	init();
	set_data(p_data);
}*/

bool dsp_lowpass::set_data(const dsp_preset & p_data)
{
	t_dsp_lpf_params params;
	if (!params.set_data(p_data)) return false;
	params.get_lpfconfig(cfg_);
	return true;
}

dsp_lowpass::~dsp_lowpass()
{
	lpf_state_ = LPF_close(lpf_state_); //lpf_state_ == NULL
	delete[] buffer_;
	buffer_ = NULL;
}

void dsp_lowpass::on_endoftrack(abort_callback & p_abort) { flushwrite(); }

void dsp_lowpass::on_endofplayback(abort_callback & p_abort) { flushwrite(); }

void dsp_lowpass::reinit(unsigned sample_rate, unsigned channel_count)
{
	int lpf_error = 0;
	LPF_config c = {sample_rate, cfg_.lp_freq, double(cfg_.tbandw10)/10.0, cfg_.phase};

    if (buf_channels_ < channel_count)
	{
		buf_channels_ = channel_count;
		delete[] buffer_;
		buffer_ = new audio_sample[buf_samples*channel_count];
		//if (buffer_ == NULL) { console::error("Resampler: new[] failed"); } //new throws exception
	}

	lpf_state_ = LPF_init(&c, channel_count, &lpf_error);
	if (lpf_state_ == NULL) { console::error(LPF_strerror(lpf_error)); }
	channel_count_ = channel_count; sample_rate_ = sample_rate;
	in_samples_accum_ = out_samples_gen_accum_ = 0;
}

bool dsp_lowpass::on_chunk(audio_chunk * chunk, abort_callback & p_abort)
{
   	size_t in_samples_used = 0, out_samples_gen = 0;
	int lpf_error = 0;

	unsigned channel_count = chunk->get_channels();
	unsigned channel_config = chunk->get_channel_config();
	unsigned sample_rate = chunk->get_sample_rate();
	t_size sample_count = chunk->get_sample_count();
	audio_sample * current = chunk->get_data();

	if (lpf_state_ == NULL) //uninitialized state
	{
		if (cfg_.cannot_lowpass(sample_rate)) return true;
		reinit(sample_rate, channel_count);
	}
	else
	{
		if ((channel_count_ != channel_count) || (sample_rate_ != sample_rate))
		{   // number of channels or samplerate has changed - reinitialize
			flushwrite();
			if (cfg_.cannot_lowpass(sample_rate)) return true;
			reinit(sample_rate, channel_count);
		}
	}
	/*if (cfg_.cannot_lowpass(sample_rate_))
		return true;*/

	in_samples_accum_ += sample_count;
	while(1)
	{
		lpf_error = LPF_flow(lpf_state_, current, buffer_, sample_count, buf_samples, &in_samples_used, &out_samples_gen);
		if (lpf_error) { console::error(LPF_strerror(lpf_error)); break; }

		out_samples_gen_accum_ += out_samples_gen;
		while (in_samples_accum_ > sample_rate_ && out_samples_gen_accum_ > sample_rate_)
		{
			in_samples_accum_ -= sample_rate_;
			out_samples_gen_accum_ -= sample_rate_;
		}
			
		//if (out_samples_gen == 0) break; //is it incorrect?
		if (out_samples_gen != 0)
		{
			audio_chunk * out = insert_chunk(out_samples_gen);
			out->set_data(buffer_, out_samples_gen, channel_count_, sample_rate_, channel_config);
		}
		current += in_samples_used * channel_count_;
		sample_count -= in_samples_used;
		if (sample_count == 0) break;
	}

	return false;
}

void dsp_lowpass::flush()
{
	if (!lpf_state_) return;
	lpf_state_ = LPF_close(lpf_state_); //lpf_state_ == NULL
	in_samples_accum_ = out_samples_gen_accum_ = 0;
	return;
}

void dsp_lowpass::flushwrite()
{
   	size_t out_samples_gen = 0;
   	int lpf_error = 0;

    if (!lpf_state_) return;
		
	while(1)
	{
		lpf_error = LPF_drain(lpf_state_, buffer_, buf_samples, &out_samples_gen);
		if (lpf_error) { console::error(LPF_strerror(lpf_error)); break; }
		if (out_samples_gen == 0) break;
		out_samples_gen_accum_ += out_samples_gen;
			
		audio_chunk * out = insert_chunk(out_samples_gen);
		out->set_data(buffer_, out_samples_gen, channel_count_, sample_rate_);
	}
	in_samples_accum_ = out_samples_gen_accum_ = 0;
	lpf_state_ = LPF_close(lpf_state_);
	return;
}
	
double dsp_lowpass::get_latency()
{
	if (sample_rate_)
		return double(in_samples_accum_)/double(sample_rate_) - double(out_samples_gen_accum_)/double(sample_rate_);
	else return 0;
}

bool dsp_lowpass::g_get_default_preset(dsp_preset & p_out)
{
	t_dsp_lpf_params().get_data(p_out);
	return true;
}

void dsp_lowpass::g_show_config_popup(const dsp_preset & p_data,HWND p_parent, dsp_preset_edit_callback & p_callback)
{
	t_dsp_lpf_params params;
	if (params.set_data(p_data))
	{
		dialog_dsp_lpf dlg(p_data, p_callback);
		dlg.DoModal(p_parent);
		/*if(dlg.DoModal(p_parent) == IDOK) { params.get_data(p_data); }*/
	}
}
