/* Effect: sinc filters     Copyright (c) 2008-9 robs@users.sourceforge.net
/* Abstract effect: dft filter     Copyright (c) 2008 robs@users.sourceforge.net
/* mix by lvqcl (c) 2009-2011
 *
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 2.1 of the License, or (at
 * your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 */

#include "sox_i.h"
#include "fft4g.h"

#define  FIFO_SIZE_T int
#include "fifo.h"

typedef struct {
  int        dft_length, num_taps, post_peak;
  double     * coefs;
  thread_fft_cache info; /*added for thread safety*/
} dft_filter_t;

typedef struct {
  size_t     samples_in, samples_out;
  fifo_t     input_fifo, output_fifo;
  dft_filter_t * filter_ptr;
} priv_t;

#include "lowpass.h"

typedef struct {
  double             att, beta, phase, Fc0, Fc1, tbw0, tbw1;
  int                num_taps[2];
  sox_bool           round;
} sinc_priv_t;


/****************************************************************************************************************/

static void invert(double * h, int n)
{
  int i;
  for (i = 0; i < n; ++i)
    h[i] = -h[i];
  h[(n - 1) / 2] += 1;
}

static double * lpf(double Fn, double Fc, double tbw, int * num_taps, double att, double * beta, sox_bool round)
{
  if ((Fc /= Fn) <= 0 || Fc >= 1) {
    *num_taps = 0;
    return NULL;
  }
  att = att? att : 120;
  *beta = *beta < 0? lsx_kaiser_beta(att) : *beta;
  if (!*num_taps) {
    int n = lsx_lpf_num_taps(att, (tbw? tbw / Fn : .05) * .5, 0);
    *num_taps = range_limit(n, 11, 32767);
    if (round)
      *num_taps = 1 + 2 * (int)((int)((*num_taps / 2) * Fc + .5) / Fc + .5);
  }
  return lsx_make_lpf(*num_taps |= 1, Fc, *beta, 1., sox_false);
}

static void lsx_set_dft_filter(dft_filter_t *f, double *h, int n, int post_peak)
{
  int i;
  f->num_taps = n;
  f->post_peak = post_peak;
  f->dft_length = lsx_set_dft_length(f->num_taps);
  f->coefs = (double*)lsx_aligned_calloc(f->dft_length, sizeof(*f->coefs));
  for (i = 0; i < f->num_taps; ++i)
    f->coefs[(i + f->dft_length - f->num_taps + 1) & (f->dft_length - 1)] = h[i] / f->dft_length * 2;
  lsx_safe_rdft_generic(f->dft_length, 1, f->coefs, &f->info);
  lsx_free(h);
}

static void filter(priv_t * p)
{
  int i, num_in = max(0, fifo_occupancy(&p->input_fifo));
  dft_filter_t * f = p->filter_ptr;
  int const overlap = f->num_taps - 1;
  double * output;

  while (num_in >= f->dft_length) {
    double const * input = (double const*)fifo_read_ptr(&p->input_fifo);
    fifo_read(&p->input_fifo, f->dft_length - overlap, NULL);
    num_in -= f->dft_length - overlap;

    output = (double*)fifo_reserve_aligned(&p->output_fifo, f->dft_length);
    fifo_trim_by(&p->output_fifo, overlap);
    memcpy(output, input, f->dft_length * sizeof(*output));

    lsx_safe_rdft_generic(f->dft_length, 1, output, &f->info);
    output[0] *= f->coefs[0];
    output[1] *= f->coefs[1];
    for (i = 2; i < f->dft_length; i += 2) {
      double tmp = output[i];
      output[i  ] = f->coefs[i  ] * tmp - f->coefs[i+1] * output[i+1];
      output[i+1] = f->coefs[i+1] * tmp + f->coefs[i  ] * output[i+1];
    }
	lsx_safe_rdft_generic(f->dft_length, -1, output, &f->info);
  }
}

/****************************************************************************************************************/

static int start(sinc_priv_t * p, priv_t * pp, int rate)
{
  dft_filter_t * f = pp->filter_ptr;

  if (!f->num_taps)
  {
    double Fn = rate * .5;
    double * h[2];
    int i, n, post_peak, longer;

    if (p->Fc0 >= Fn || p->Fc1 >= Fn) {
      return SOX_EOF;
    }
    h[0] = lpf(Fn, p->Fc0, p->tbw0, &p->num_taps[0], p->att, &p->beta,p->round);
    h[1] = lpf(Fn, p->Fc1, p->tbw1, &p->num_taps[1], p->att, &p->beta,p->round);
    if (h[0])
      invert(h[0], p->num_taps[0]);

    longer = p->num_taps[1] > p->num_taps[0];
    n = p->num_taps[longer];
    if (h[0] && h[1]) {
      for (i = 0; i < p->num_taps[!longer]; ++i)
        h[longer][i + (n - p->num_taps[!longer])/2] += h[!longer][i];

      if (p->Fc0 < p->Fc1)
        invert(h[longer], n);

      lsx_free(h[!longer]);
    }
    if (p->phase != 50)
		lsx_fir_to_phase(&h[longer], &n, &post_peak, p->phase, &(f->info));
    else post_peak = n >> 1;

    lsx_set_dft_filter(f, h[longer], n, post_peak);
  }

  //return lsx_dft_filter_effect_fn()->start(effp);
  fifo_create(&pp->input_fifo, (int)sizeof(double));
  memset(fifo_reserve(&pp->input_fifo, f->post_peak), 0, sizeof(double) * f->post_peak);
  fifo_create(&pp->output_fifo, (int)sizeof(double));

  return SOX_SUCCESS;
}

static int stop(priv_t * p)
{
  fifo_delete(&p->input_fifo);
  fifo_delete(&p->output_fifo);
  p->filter_ptr = NULL;
  return SOX_SUCCESS;
}

/****************************************************************************************************************/

/*------------------------------- Wrapper for foobar2000 --------------------------------*/
/* Copyright (c) 2008-09 lvqcl.
 *
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 2.1 of the License, or (at
 * your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 */ 

typedef struct LPF_internal_tag
{	
	int nchannels;
	priv_t* dftparam;
	dft_filter_t filter;
} LPF_internal; 

//static float _2_32 = 1.0/65536.0/65536.0;//0.00000000023283064365386962890625

LPF_handle* LPF_init_generic(const LPF_config* config, int nchannels, int *error) /*create+start*/
{
	LPF_internal* p;
	sinc_priv_t sincparam;
	int ret;
	int i;

	if (config->lp_freq >= config->samplerate/2)
	{
		if (error) *error = LPF_OK;
		return NULL;
	}

	sincparam.att = 150;//dB
	sincparam.beta = -1.0;
	sincparam.Fc0 = 0.0;
	sincparam.tbw0 = 0.0;
	sincparam.Fc1 = config->lp_freq;
	if (sincparam.Fc1 == 0.0)
		sincparam.Fc1 = 1.0;
	//sincparam.tbw1 = ((double)config->bandwidth)/100.0 * (config->samplerate/2.0);
	sincparam.tbw1 = ((double)config->bandwidth)/100.0 * config->lp_freq;
	sincparam.phase = config->phase;
	sincparam.round = sox_false;
	sincparam.num_taps[0] = 0;
	sincparam.num_taps[1] = 0;

	p = (LPF_internal*)lsx_calloc(1, sizeof (LPF_internal));
	p->dftparam = (priv_t*) lsx_calloc(nchannels, sizeof(priv_t));
	init_fft_cache(&p->filter.info);
	p->nchannels = nchannels;
	for(i=0; i<nchannels; i++)
	{
		priv_t* dp = &p->dftparam[i];
		dp->filter_ptr = &p->filter;
		ret = start(&sincparam, dp, config->samplerate);
		if(error && ret) *error = ret;
	}

	return (LPF_handle*)p; 
}

#define NSAMP 1000000

int LPF_flow_generic(LPF_handle *h, const fb_sample_t* ibuf, fb_sample_t* obuf, size_t isamp, size_t osamp, size_t* i_used, size_t* o_done) //flow
{
	LPF_internal* p = (LPF_internal*)h;
	size_t n;
	int i, nchannels;
	sample_t const * s;
	size_t odone;
	if(p == NULL) return LPF_NULLHANDLE;
	
	nchannels = p->nchannels;
	for(i=0; i<nchannels; i++)
	{
		priv_t* dp = &p->dftparam[i];
		odone = min(osamp, (size_t)fifo_occupancy(&dp->output_fifo));
		s = (sample_t const*)fifo_read(&dp->output_fifo, (int)odone, NULL);
		for (n = 0; n < odone; ++n)
				obuf[n*nchannels+i] = (fb_sample_t) s[n];

		dp->samples_out += odone;

		if (isamp && odone < osamp)
		{
			double * t = (double*)fifo_write(&dp->input_fifo, (int)isamp, NULL);
			dp->samples_in += (int)isamp;

			//for (n = isamp; n; --n) t[isamp-n] = (sample_t) ibuf[(isamp-n)*nchannels+i];
			for (n = 0; n<isamp; n++) t[n] = (sample_t) ibuf[n*nchannels+i];

			filter(dp);
			*i_used = isamp;
		}
		else
			*i_used = 0;
		
		*o_done = odone;

		while ((dp->samples_in >= NSAMP) && (dp->samples_out >= NSAMP))
		{
			dp->samples_in -= NSAMP;
			dp->samples_out -= NSAMP;
		}

	}
	return SOX_SUCCESS;
}

int LPF_drain_generic(LPF_handle *h, fb_sample_t* obuf, size_t osamp, size_t* ogen) //drain
{
	LPF_internal* p = (LPF_internal*)h;
	size_t igen = 0;
	int i;
	double* buff;
	
	if(p == NULL) return LPF_NULLHANDLE;
	buff = (double*)lsx_calloc(1024, sizeof(*buff));

	for(i=0; i<p->nchannels; i++)
	{
		priv_t* dp = &p->dftparam[i];

		if (dp->samples_in > dp->samples_out)
		{
			int remaining = dp->samples_in - dp->samples_out;
			while (fifo_occupancy(&dp->output_fifo) < remaining)
			{
				fifo_write(&dp->input_fifo, 1024, buff);
				dp->samples_in += 1024;
				filter(dp);
			}
			fifo_trim_to(&dp->output_fifo, remaining);
			dp->samples_in = 0;
		}
	}
	lsx_free(buff);

	return LPF_flow(h, NULL, obuf, 0, osamp, &igen, ogen);
}

LPF_handle* LPF_close_generic(LPF_handle *h) /*stop*/
{
	LPF_internal* p = (LPF_internal*)h;
	int i;
	if(p == NULL) return NULL;

	for(i=0; i<p->nchannels; i++)
		stop(&p->dftparam[i]);

	lsx_aligned_free(p->filter.coefs);

	free_fft_cache(&p->filter.info);
	lsx_free(p->dftparam);
	lsx_free(p);
	return NULL;
}

void lpf_init0_generic()
{
	LPF_init = LPF_init_generic;
	LPF_close = LPF_close_generic;
	LPF_flow = LPF_flow_generic;
	LPF_drain = LPF_drain_generic;
}

const char* LPF_strerror(int error)
{
	switch(error)
	{
	case LPF_OK:
		return "OK";
	case LPF_ENOMEM:
		return "Not enough memory";
	case LPF_INTERNAL:
		return "Internal error";
	case LPF_NULLHANDLE:
		return "NULL handle";
	case LPF_SOXERROR:
	case LPF_EOF:
		return "Error in SoX functions";
	default:
		return "Other error";
	}
}
